from aiohttp import web

from .utils import issubclass_py37fix


@web.middleware
async def validation_middleware(request: web.Request, handler) -> web.Response:
    """
    Validation middleware for aiohttp web app

    Usage:

    .. code-block:: python

        app.middlewares.append(validation_middleware)


    """
    orig_handler = request.match_info.handler
    if not hasattr(orig_handler, "__schemas__"):
        if not issubclass_py37fix(orig_handler, web.View):
            return await handler(request)
        sub_handler = getattr(orig_handler, request.method.lower(), None)
        if sub_handler is None:
            return await handler(request)
        if not hasattr(sub_handler, "__schemas__"):
            return await handler(request)
        schemas = sub_handler.__schemas__
    else:
        schemas = orig_handler.__schemas__  # type: ignore
    result = {}
    for schema in schemas:
        data = await request.app["_apispec_parser"].parse(
            schema["schema"],
            request,
            location=schema["location"],
        )
        if schema["put_into"]:
            request[schema["put_into"]] = data
        elif data:
            try:
                result.update(data)
            except (ValueError, TypeError):
                result = data
                break
    request[request.app["_apispec_request_data_name"]] = result
    return await handler(request)



import logging

import schema
from aiohttp import web

logger = logging.getLogger(__name__)

routes = web.RouteTableDef()

CLIENT_SCHEMA = schema.Schema(
    schema.Or(
        {
            "type": "subscribe",
            "token": str,
            "data": {"subscription_type": str, "element_id": int},
        },
        {
            "type": "unsubscribe",
            "token": str,
            "data": {"subscription_type": str, "element_id": int},
        },
    )
)

SERVER_SCHEMA = schema.Schema(
    schema.Or(
        {"type": "error", "data": {"code": int, "msg": str}},
        {
            "type": "notification",
            "data": {
                "subscription_type": str,
                "element_id": int,
            },
        },
        {
            "type": "subscribe_success",
            "data": {
                "subscription_type": str,
                "element_id": int,
            },
        },
        {
            "type": "unsubscribe_success",
            "data": {
                "subscription_type": str,
                "element_id": int,
            },
        },
    )
)


def make_error_msg(code: int, msg: str) -> dict:
    return {"type": "error", "data": {"code": code, "msg": msg}}



import copy
from typing import Awaitable, Callable

from aiohttp import web
from aiohttp.hdrs import METH_ALL, METH_ANY
from apispec import APISpec
from apispec.core import VALID_METHODS_OPENAPI_V2
from apispec.ext.marshmallow import MarshmallowPlugin, common
from webargs.aiohttpparser import parser

from .utils import get_path, get_path_keys, issubclass_py37fix

_AiohttpView = Callable[[web.Request], Awaitable[web.StreamResponse]]

VALID_RESPONSE_FIELDS = {"description", "headers", "examples"}

NAME_SWAGGER_SPEC = "swagger.spec"


def resolver(schema):
    schema_instance = common.resolve_schema_instance(schema)
    prefix = "Partial-" if schema_instance.partial else ""
    schema_cls = common.resolve_schema_cls(schema)
    name = prefix + schema_cls.__name__
    if name.endswith("Schema"):
        return name[:-6] or name
    return name


class AiohttpApiSpec:
    def __init__(
        self,
        url="/api/docs/swagger.json",
        app=None,
        request_data_name="data",
        error_callback=None,
        in_place=False,
        prefix="",
        schema_name_resolver=resolver,
        **kwargs
    ):

        self.plugin = MarshmallowPlugin(schema_name_resolver=schema_name_resolver)
        self.spec = APISpec(plugins=(self.plugin,), openapi_version="2.0", **kwargs)

        self.url = url
        self._registered = False
        self._request_data_name = request_data_name
        self.error_callback = error_callback
        self.prefix = prefix
        self._index_page = None
        if app is not None:
            self.register(app, in_place)

    def swagger_dict(self):
        """Returns swagger spec representation in JSON format"""
        return self.spec.to_dict()

    def register(self, app: web.Application, in_place: bool = False):
        """Creates spec based on registered app routes and registers needed view"""
        if self._registered is True:
            return None

        app["_apispec_request_data_name"] = self._request_data_name

        if self.error_callback:
            parser.error_callback = self.error_callback
        app["_apispec_parser"] = parser

        if in_place:
            self._register(app)
        else:

            async def doc_routes(app_):
                self._register(app_)

            app.on_startup.append(doc_routes)

        self._registered = True

        if self.url is not None:

            async def swagger_handler(request):
                return web.json_response(request.app["swagger_dict"])

            route_url = self.url
            if not self.url.startswith("/"):
                route_url = "/{}".format(self.url)
            app.router.add_route(
                "GET", route_url, swagger_handler, name=NAME_SWAGGER_SPEC
            )

    def _register(self, app: web.Application):
        for route in app.router.routes():
            if issubclass_py37fix(route.handler, web.View) and route.method == METH_ANY:
                for attr in dir(route.handler):
                    if attr.upper() in METH_ALL:
                        view = getattr(route.handler, attr)
                        method = attr
                        self._register_route(route, method, view)
            else:
                method = route.method.lower()
                view = route.handler
                self._register_route(route, method, view)
        app["swagger_dict"] = self.swagger_dict()

    def _register_route(
        self, route: web.AbstractRoute, method: str, view: _AiohttpView
    ):

        if not hasattr(view, "__apispec__"):
            return None

        url_path = get_path(route)
        if not url_path:
            return None

        self._update_paths(view.__apispec__, method, self.prefix + url_path)  # type: ignore

    def _update_paths(self, data: dict, method: str, url_path: str):
        if method not in VALID_METHODS_OPENAPI_V2:
            return None
        for schema in data.pop("schemas", []):
            parameters = self.plugin.converter.schema2parameters(
                schema["schema"], **schema["options"]
            )
            self._add_examples(schema["schema"], parameters, schema["example"])
            data["parameters"].extend(parameters)

        existing = [p["name"] for p in data["parameters"] if p["in"] == "path"]
        data["parameters"].extend(
            {"in": "path", "name": path_key, "required": True, "type": "string"}
            for path_key in get_path_keys(url_path)
            if path_key not in existing
        )

        if "responses" in data:
            responses = {}
            for code, actual_params in data["responses"].items():
                if "schema" in actual_params:
                    raw_parameters = self.plugin.converter.schema2parameters(
                        actual_params["schema"],
                        location="json",
                        required=actual_params.get("required", False),
                    )[0]
                    updated_params = {
                        k: v
                        for k, v in raw_parameters.items()
                        if k in VALID_RESPONSE_FIELDS
                    }
                    updated_params["schema"] = actual_params["schema"]
                    for extra_info in ("description", "headers", "examples"):
                        if extra_info in actual_params:
                            updated_params[extra_info] = actual_params[extra_info]
                    responses[code] = updated_params
                else:
                    responses[code] = actual_params
            data["responses"] = responses

        operations = copy.deepcopy(data)
        self.spec.path(path=url_path, operations={method: operations})

    def _add_examples(self, ref_schema, endpoint_schema, example):
        def add_to_endpoint_or_ref():
            if add_to_refs:
                self.spec.components._schemas[name]["example"] = example
            else:
                endpoint_schema[0]["schema"]["allOf"] = [
                    endpoint_schema[0]["schema"].pop("$ref")
                ]
                endpoint_schema[0]["schema"]["example"] = example

        if not example:
            return
        schema_instance = common.resolve_schema_instance(ref_schema)
        name = self.plugin.converter.schema_name_resolver(schema_instance)
        add_to_refs = example.pop("add_to_refs")
        if self.spec.components.openapi_version.major < 3:
            if name and name in self.spec.components._schemas:
                add_to_endpoint_or_ref()
        else:
            add_to_endpoint_or_ref()



"""
Taken from https://github.com/maximdanilchenko/aiohttp-apispec

aiohttp-apispec is release under the MIT License

Copyright (c) 2017 Maksim Danilchenko
"""

from .setup import setup_aiohttp_apispec
from .decorators import (
    docs,
    request_schema,
    match_info_schema,
    querystring_schema,
    form_schema,
    json_schema,
    headers_schema,
    cookies_schema,
    response_schema,
    use_kwargs,
    marshal_with,
)
from .middlewares import validation_middleware

__all__ = [
    # setup
    "setup_aiohttp_apispec",
    # decorators
    "docs",
    "request_schema",
    "match_info_schema",
    "querystring_schema",
    "form_schema",
    "json_schema",
    "headers_schema",
    "cookies_schema",
    "response_schema",
    "use_kwargs",
    "marshal_with",
    # middleware
    "validation_middleware",
]



"""
Provides the SubCommand abstract base class
"""
import abc

from abrechnung.config import Config


class SubCommand(abc.ABC):
    """
    Abstract base class for subcommands.

    The constructor receives the config dictionary as an argument,
    plus all arguments that were added by argparse_register() as keyword
    arguments.
    """

    @staticmethod
    def argparse_register(subparser):
        """
        Implement this method to add your own argparse arguments
        to this subparser.

        No arguments are added by default
        """
        del subparser  # unused

    @staticmethod
    def argparse_validate(args, error_cb):
        """
        Implement this method to add your own custom validation
        of arguparse arguments.

        This will be called immediately after the arguments have been parsed;
        calling error_cb with a str argument causes argparse to exit with that
        error message.

        No checking is performed by default
        """
        del args, error_cb  # unused

    @abc.abstractmethod
    def __init__(self, config: Config, **kwargs):
        pass

    @abc.abstractmethod
    async def run(self):
        """
        This is the entry point that will be called after the program has been
        initialized.
        """



import functools
import json
from datetime import datetime, date
from typing import Optional, Any
from uuid import UUID

import asyncpg
from aiohttp import web
from aiohttp.helpers import sentinel
from aiohttp.typedefs import LooseHeaders
from marshmallow import ValidationError

from abrechnung.application import NotFoundError, InvalidCommand


def validate(request_schema: type):
    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args):
            try:
                req_body = await request.json()
            except (json.decoder.JSONDecodeError, TypeError):
                raise web.HTTPBadRequest(
                    reason="Request is malformed; could not decode JSON object."
                )

            try:
                request_schema().load(req_body)
            except ValidationError as e:
                raise web.HTTPBadRequest(
                    reason=f"Request is invalid; there are validation errors: {e.messages}"
                )

            view_args = request, req_body
            response = await func(*view_args)

            return response

        return wrapped

    return wrapper


@web.middleware
async def error_middleware(request, handler):
    try:
        response = await handler(request)
        if response.status < 400:
            return response
        message = response.message
        status = response.status
    except (
        asyncpg.DataError,
        asyncpg.RaiseError,
        asyncpg.IntegrityConstraintViolationError,
        InvalidCommand,
    ) as ex:
        # catch underlying bad query inputs to catch schema violations as bad requests
        status = web.HTTPBadRequest.status_code
        message = str(ex)
    except NotFoundError as ex:
        # catch underlying bad query inputs to catch schema violations as bad requests
        status = web.HTTPNotFound.status_code
        message = str(ex)
    except PermissionError as ex:
        # catch underlying permission errors
        status = web.HTTPForbidden.status_code
        message = str(ex)
    except web.HTTPException as ex:
        message = ex.reason
        status = ex.status
    # except Exception as ex:  # TODO: deliberate whether this is such a good idea, but it guarantees json responses
    #     status = web.HTTPInternalServerError.status_code
    #     message = str(ex)

    return web.json_response({"code": status, "msg": message}, status=status)


def encode_json(obj):
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    if isinstance(obj, UUID):
        return str(obj)

    raise TypeError(f"Type {type(obj)} is not serializable")


def json_response(
    data: Any = sentinel,
    *,
    text: Optional[str] = None,
    body: Optional[bytes] = None,
    status: int = 200,
    reason: Optional[str] = None,
    headers: Optional[LooseHeaders] = None,
    content_type: str = "application/json",
):
    return web.json_response(
        data,
        text=text,
        body=body,
        status=status,
        reason=reason,
        headers=headers,
        content_type=content_type,
        dumps=lambda x: json.dumps(x, default=encode_json),
    )



import asyncio
from dataclasses import dataclass
from typing import Optional

from aiosmtpd import smtp
from aiosmtpd.controller import Controller

from abrechnung.application.users import UserService
from abrechnung.config import Config
from abrechnung.mailer import Mailer
from tests import AsyncTestCase, get_test_db_config


@dataclass
class Mail:
    peer: str
    from_email: str
    rcpttos: str
    data: str
    mail_options: Optional[str]
    rcpt_options: Optional[str]


class DummySMTPHandler:
    def __init__(self):
        self.mail_queue = asyncio.Queue()

    async def handle_RCPT(
        self, server, session, envelope: smtp.Envelope, address: str, rcpt_options
    ):
        del server, session, rcpt_options  # unused
        envelope.rcpt_tos.append(address)
        return "250 OK"

    async def handle_DATA(self, server, session, envelope: smtp.Envelope):
        del server, session  # unused
        await self.mail_queue.put(envelope)
        return "250 Message accepted for delivery"


class MailerTest(AsyncTestCase):
    async def setUpAsync(self) -> None:
        await super().setUpAsync()
        self.smtp_handler = DummySMTPHandler()
        self.smtp = Controller(self.smtp_handler)
        self.smtp.start()

        self.mailer_config = Config.from_dict(
            {
                "api": {"enable_registration": True},
                "email": {
                    "host": self.smtp.hostname,
                    "port": self.smtp.port,
                    "address": "abrechnung@stusta.de",
                },
                "database": get_test_db_config(),
                "service": {
                    "url": "https://abrechnung.example.lol",
                    "name": "Test Abrechnung",
                },
            }
        )
        self.mailer = Mailer(config=self.mailer_config)

        self.mailer_task = asyncio.create_task(self.mailer.run())

        self.user_service = UserService(db_pool=self.db_pool, config=self.mailer_config)

    async def tearDownAsync(self) -> None:
        await super().tearDownAsync()
        self.mailer_task.cancel()
        self.smtp.stop()

    async def test_registration_mail_delivery(self):
        user_email = "user@email.com"
        await self.user_service.register_user("user1", user_email, "password")
        await asyncio.sleep(0.5)

        mail: smtp.Envelope = self.smtp_handler.mail_queue.get_nowait()
        self.assertIsNotNone(mail)
        self.assertIn(user_email, mail.rcpt_tos)
        self.assertIn(
            "[Test Abrechnung] Confirm user account", mail.content.decode("utf-8")
        )

    async def test_email_change_mail_delivery(self):
        user_email = "user@email.com"
        new_email = "new_email@email.com"
        user_id, password = await self._create_test_user(
            username="user", email=user_email
        )
        await self.user_service.request_email_change(
            user_id=user_id, password=password, email=new_email
        )

        await asyncio.sleep(0.5)
        mail1: smtp.Envelope = self.smtp_handler.mail_queue.get_nowait()
        self.assertIsNotNone(mail1)
        mail2: smtp.Envelope = self.smtp_handler.mail_queue.get_nowait()
        self.assertIsNotNone(mail2)
        self.assertTrue(user_email in mail1.rcpt_tos or mail2.rcpt_tos)
        self.assertTrue(new_email in mail1.rcpt_tos or mail2.rcpt_tos)
        self.assertIn("[Test Abrechnung] Change email", mail1.content.decode("utf-8"))
        self.assertIn("[Test Abrechnung] Change email", mail2.content.decode("utf-8"))

    async def test_password_reset_mail_delivery(self):
        user_email = "user@email.com"
        await self._create_test_user(username="user", email=user_email)
        await self.user_service.request_password_recovery(email=user_email)

        await asyncio.sleep(0.5)
        mail: smtp.Envelope = self.smtp_handler.mail_queue.get_nowait()
        self.assertIsNotNone(mail)
        self.assertIn(user_email, mail.rcpt_tos)
        self.assertIn("[Test Abrechnung] Reset password", mail.content.decode("utf-8"))



from functools import partial
import copy


def request_schema(
    schema, location=None, put_into=None, example=None, add_to_refs=False, **kwargs
):
    """
    Add request info into the swagger spec and
    prepare injection keyword arguments from the specified
    webargs arguments into the decorated view function in
    request['data'] for validation_middleware validation middleware.

    Usage:

    .. code-block:: python

        from aiohttp import web
        from marshmallow import Schema, fields


        class RequestSchema(Schema):
            id = fields.Int()
            name = fields.Str(description='name')

        @request_schema(RequestSchema(strict=True))
        async def index(request):
            # aiohttp_apispec_middleware should be used for it
            data = request['data']
            return web.json_response({'name': data['name'],
                                      'id': data['id']})

    :param schema: :class:`Schema <marshmallow.Schema>` class or instance
    :param locations: Default request locations to parse
    :param put_into: name of the key in Request object
                     where validated data will be placed.
                     If None (by default) default key will be used
    :param dict example: Adding example for current schema
    :param bool add_to_refs: Working only if example not None,
                             if True, add example for ref schema.
                             Otherwise add example to endpoint.
                             Default False
    """
    if callable(schema):
        schema = schema()

    # Compatability with old versions should be dropped,
    # multiple locations are no longer supported by a single call
    # so therefore **locations should never be used

    options = {"required": kwargs.pop("required", False)}
    # to support apispec >=4 need to rename default_in
    if location:
        options["location"] = location
    elif "default_in" not in options:
        options["location"] = "body"

    def wrapper(func):
        if not hasattr(func, "__apispec__"):
            func.__apispec__ = {"schemas": [], "responses": {}, "parameters": []}
            func.__schemas__ = []

        _example = copy.copy(example) or {}
        if _example:
            _example["add_to_refs"] = add_to_refs
        func.__apispec__["schemas"].append(
            {"schema": schema, "options": options, "example": _example}
        )
        # TODO: Remove this block?
        if location and "body" in location:
            body_schema_exists = (
                "body" in func_schema["location"] for func_schema in func.__schemas__
            )
            if any(body_schema_exists):
                raise RuntimeError("Multiple body parameters are not allowed")

        func.__schemas__.append(
            {"schema": schema, "location": location, "put_into": put_into}
        )

        return func

    return wrapper


# For backward compatibility
use_kwargs = request_schema

# Decorators for specific request data validations (shortenings)
match_info_schema = partial(
    request_schema, location="match_info", put_into="match_info"
)
querystring_schema = partial(
    request_schema, location="querystring", put_into="querystring"
)
form_schema = partial(request_schema, location="form", put_into="form")
json_schema = partial(request_schema, location="json", put_into="json")
headers_schema = partial(request_schema, location="headers", put_into="headers")
cookies_schema = partial(request_schema, location="cookies", put_into="cookies")



import asyncio
import logging
import os
import re
import signal
import sys
import termios

# the vt100 CONTROL SEQUENCE INTRODUCER
from datetime import datetime, timedelta, timezone

CSI = "\N{ESCAPE}["


def SGR(code=""):
    """
    Returns a SELECT GRAPHIC RENDITION code sequence
    for the given code.
    See https://en.wikipedia.org/wiki/ANSI_escape_code
    """
    return f"{CSI}{code}m"


BOLD = SGR(1)
RED = SGR(31)
NORMAL = SGR()

postgres_timestamp_format = re.compile(
    r"(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})T(?P<hour>\d{2}):(?P<minute>\d{2}):(?P<second>\d{2})"
    r"(\.(?P<subseconds>\d+))?(?P<tzsign>[-+])(?P<tzhours>\d{2}):(?P<tzminutes>\d{2})"
)


def parse_postgres_datetime(dt: str) -> datetime:
    if m := postgres_timestamp_format.match(dt):
        subseconds = m.group("subseconds")
        micros = 0
        if len(subseconds) > 3:
            millis = int(subseconds[:3])
            micros = int(subseconds[3:].ljust(3, "0"))
        else:
            millis = int(subseconds.ljust(3, "0"))

        tzsign = -1 if m.group("tzsign") == "-" else 1
        tzdelta = timedelta(
            hours=int(m.group("tzhours")), minutes=int(m.group("tzminutes"))
        )
        tz = timezone(tzsign * tzdelta)

        return datetime(
            year=int(m.group("year")),
            month=int(m.group("month")),
            day=int(m.group("day")),
            hour=int(m.group("hour")),
            minute=int(m.group("minute")),
            second=int(m.group("second")),
            microsecond=millis * 1000 + micros,
            tzinfo=tz,
        )

    raise ValueError("invalid format")


def format_error(text):
    """
    Formats an error text for printing in a terminal
    """
    return f"\N{PILE OF POO} {RED}{BOLD}{text}{NORMAL}"


def log_setup(setting, default=1):
    """
    Perform setup for the logger.
    Run before any logging.log thingy is called.

    if setting is 0: the default is used, which is WARNING.
    else: setting + default is used.
    """

    levels = (
        logging.ERROR,
        logging.WARNING,
        logging.INFO,
        logging.DEBUG,
        logging.NOTSET,
    )

    factor = clamp(default + setting, 0, len(levels) - 1)
    level = levels[factor]

    logging.basicConfig(level=level, format="[%(asctime)s] %(message)s")
    logging.captureWarnings(True)


def clamp(number, smallest, largest):
    """return number but limit it to the inclusive given value range"""
    return max(smallest, min(number, largest))


async def run_as_fg_process(args, **kwargs):
    """
    the "correct" way of spawning a new subprocess:
    signals like C-c must only go
    to the child process, and not to this python.

    the args are the same as subprocess.Popen

    returns Popen().wait() value

    Some side-info about "how ctrl-c works":
    https://unix.stackexchange.com/a/149756/1321

    fun fact: this function took a whole night
              to be figured out.
    """

    old_pgrp = os.tcgetpgrp(sys.stdin.fileno())
    old_attr = termios.tcgetattr(sys.stdin.fileno())

    user_preexec_fn = kwargs.pop("preexec_fn", None)

    def new_pgid():
        if user_preexec_fn:
            user_preexec_fn()

        # set a new process group id
        os.setpgid(os.getpid(), os.getpid())

        # generally, the child process should stop itself
        # before exec so the parent can set its new pgid.
        # (setting pgid has to be done before the child execs).
        # however, Python 'guarantee' that `preexec_fn`
        # is run before `Popen` returns.
        # this is because `Popen` waits for the closure of
        # the error relay pipe '`errpipe_write`',
        # which happens at child's exec.
        # this is also the reason the child can't stop itself
        # in Python's `Popen`, since the `Popen` call would never
        # terminate then.
        # `os.kill(os.getpid(), signal.SIGSTOP)`

    try:
        # fork the child
        child = await asyncio.create_subprocess_exec(
            *args, preexec_fn=new_pgid, **kwargs
        )

        # we can't set the process group id from the parent since the child
        # will already have exec'd. and we can't SIGSTOP it before exec,
        # see above.
        # `os.setpgid(child.pid, child.pid)`

        # set the child's process group as new foreground
        os.tcsetpgrp(sys.stdin.fileno(), child.pid)
        # revive the child,
        # because it may have been stopped due to SIGTTOU or
        # SIGTTIN when it tried using stdout/stdin
        # after setpgid was called, and before we made it
        # forward process by tcsetpgrp.
        os.kill(child.pid, signal.SIGCONT)

        # wait for the child to terminate
        ret = await child.wait()

    finally:
        # we have to mask SIGTTOU because tcsetpgrp
        # raises SIGTTOU to all current background
        # process group members (i.e. us) when switching tty's pgrp
        # it we didn't do that, we'd get SIGSTOP'd
        hdlr = signal.signal(signal.SIGTTOU, signal.SIG_IGN)
        # make us tty's foreground again
        os.tcsetpgrp(sys.stdin.fileno(), old_pgrp)
        # now restore the handler
        signal.signal(signal.SIGTTOU, hdlr)
        # restore terminal attributes
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN, old_attr)

    return ret



from .docs import docs
from .request import (
    request_schema,
    use_kwargs,  # for backward compatibility
    match_info_schema,  # request_schema with locations=["match_info"]
    querystring_schema,  # request_schema with locations=["querystring"]
    form_schema,  # request_schema with locations=["form"]
    json_schema,  # request_schema with locations=["json"]
    headers_schema,  # request_schema with locations=["headers"]
    cookies_schema,  # request_schema with locations=["cookies"]
)
from .response import (
    response_schema,
    marshal_with,  # for backward compatibility
)



from datetime import datetime, timezone
from typing import Optional

import bcrypt
from asyncpg.pool import Pool
from email_validator import validate_email, EmailNotValidError

from abrechnung.domain.users import User, Session
from . import Application, NotFoundError, InvalidCommand
from ..config import Config


class InvalidPassword(Exception):
    pass


class LoginFailed(Exception):
    pass


class UserService(Application):
    def __init__(
        self,
        db_pool: Pool,
        config: Config,
    ):
        super().__init__(db_pool=db_pool, config=config)

        self.enable_registration = self.cfg["api"]["enable_registration"]
        self.valid_email_domains = self.cfg["api"]["valid_email_domains"]

    @staticmethod
    def _hash_password(password: str) -> str:
        salt = bcrypt.gensalt()
        return bcrypt.hashpw(password=password.encode("utf-8"), salt=salt).decode()

    @staticmethod
    def _check_password(password: str, hashed_password: str) -> bool:
        return bcrypt.checkpw(
            password=password.encode("utf-8"),
            hashed_password=hashed_password.encode("utf-8"),
        )

    async def _verify_user_password(self, user_id: int, password: str) -> bool:
        async with self.db_pool.acquire() as conn:
            user = await conn.fetchrow(
                "select hashed_password, pending, deleted from usr where id = $1",
                user_id,
            )
            if user is None:
                raise NotFoundError(f"User with id {user_id} does not exist")

            if user["deleted"] or user["pending"]:
                return False

            return self._check_password(password, user["hashed_password"])

    async def is_session_token_valid(self, token: str) -> Optional[tuple[int, int]]:
        """returns the session id"""
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                row = await conn.fetchrow(
                    "select user_id, id from session where token = $1 and valid_until is null or valid_until > now()",
                    token,
                )
                if row:
                    await conn.execute(
                        "update session set last_seen = now() where token = $1", token
                    )

                return row

    async def login_user(
        self, username: str, password: str, session_name: str
    ) -> tuple[int, int, str]:
        """
        validate whether a given user can login

        If successful return the user id, a new session id and a session token
        """
        async with self.db_pool.acquire(timeout=1) as conn:
            async with conn.transaction():
                user = await conn.fetchrow(
                    "select id, hashed_password, pending, deleted from usr where username = $1 or email = $1",
                    username,
                )
                if user is None:
                    raise InvalidCommand(f"Login failed")

                if not self._check_password(password, user["hashed_password"]):
                    raise InvalidCommand(f"Login failed")

                if user["deleted"]:
                    raise InvalidCommand(f"User is not permitted to login")

                if user["pending"]:
                    raise InvalidCommand(
                        f"You need to confirm your email before logging in"
                    )

                session_token, session_id = await conn.fetchrow(
                    "insert into session (user_id, name) values ($1, $2) returning token, id",
                    user["id"],
                    session_name,
                )

                return user["id"], session_id, session_token

    async def logout_user(self, *, user_id: int, session_id: int):
        async with self.db_pool.acquire(timeout=1) as conn:
            async with conn.transaction():
                sess_id = await conn.fetchval(
                    "delete from session where id = $1 and user_id = $2 returning id",
                    session_id,
                    user_id,
                )
                if sess_id is None:
                    raise InvalidCommand(f"Already logged out")

    async def register_user(self, username: str, email: str, password: str) -> int:
        """Register a new user, returning the newly created user id and creating a pending registration entry"""
        if not self.enable_registration:
            raise PermissionError(f"User registrations are disabled on this server")

        if self.valid_email_domains is not None:
            splitted = email.split("@")
            if len(splitted) == 0:
                raise InvalidCommand(f"Invalid email {email}")

            domain = splitted[-1]
            if domain not in self.valid_email_domains:
                raise PermissionError(
                    f"Only users with emails out of the following domains are "
                    f"allowed: {self.valid_email_domains}"
                )

        try:
            valid = validate_email(email)
            email = valid.email
        except EmailNotValidError as e:
            raise InvalidCommand(str(e))

        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                hashed_password = self._hash_password(password)
                user_id = await conn.fetchval(
                    "insert into usr (username, email, hashed_password) values ($1, $2, $3) returning id",
                    username,
                    email,
                    hashed_password,
                )
                if user_id is None:
                    raise InvalidCommand(f"Registering new user failed")

                await conn.execute(
                    "insert into pending_registration (user_id) values ($1)", user_id
                )

                return user_id

    async def confirm_registration(self, token: str) -> int:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                row = await conn.fetchrow(
                    "select user_id, valid_until from pending_registration where token = $1",
                    token,
                )
                if row is None:
                    raise PermissionError(f"Invalid registration token")

                user_id = row["user_id"]
                valid_until = row["valid_until"]
                if valid_until is None or valid_until < datetime.now(tz=timezone.utc):
                    raise PermissionError(f"Invalid registration token")

                await conn.execute(
                    "delete from pending_registration where user_id = $1", user_id
                )
                await conn.execute(
                    "update usr set pending = false where id = $1", user_id
                )

                return user_id

    async def get_user(self, user_id: int) -> User:
        async with self.db_pool.acquire() as conn:
            user = await conn.fetchrow(
                "select id, email, registered_at, username, pending, deleted from usr where id = $1",
                user_id,
            )

            if user is None:
                raise NotFoundError(f"User with id {user_id} does not exist")

            rows = await conn.fetch(
                "select id, name, valid_until, last_seen from session where user_id = $1",
                user_id,
            )
            sessions = [
                Session(
                    id=row["id"],
                    name=row["name"],
                    valid_until=row["valid_until"],
                    last_seen=row["last_seen"],
                )
                for row in rows
            ]

            return User(
                id=user["id"],
                email=user["email"],
                registered_at=user["registered_at"],
                username=user["username"],
                pending=user["pending"],
                deleted=user["deleted"],
                sessions=sessions,
            )

    async def delete_session(self, user_id: int, session_id: int):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                sess_id = await conn.fetchval(
                    "delete from session where id = $1 and user_id = $2 returning id",
                    session_id,
                    user_id,
                )
                if not sess_id:
                    raise NotFoundError(f"no such session found with id {session_id}")

    async def rename_session(self, user_id: int, session_id: int, name: str):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                sess_id = await conn.fetchval(
                    "update session set name = $3 where id = $1 and user_id = $2 returning id",
                    session_id,
                    user_id,
                    name,
                )
                if not sess_id:
                    raise NotFoundError(f"no such session found with id {session_id}")

    async def change_password(self, user_id: int, old_password: str, new_password: str):
        async with self.db_pool.acquire() as conn:
            valid_pw = await self._verify_user_password(user_id, old_password)
            if not valid_pw:
                raise InvalidPassword

            hashed_password = self._hash_password(new_password)
            await conn.execute(
                "update usr set hashed_password = $1 where id = $2",
                hashed_password,
                user_id,
            )

    async def request_email_change(self, user_id: int, password: str, email: str):
        try:
            valid = validate_email(email)
            email = valid.email
        except EmailNotValidError as e:
            raise InvalidCommand(str(e))

        async with self.db_pool.acquire() as conn:
            valid_pw = await self._verify_user_password(user_id, password)
            if not valid_pw:
                raise InvalidPassword

            await conn.execute(
                "insert into pending_email_change (user_id, new_email) values ($1, $2)",
                user_id,
                email,
            )

    async def confirm_email_change(self, token: str) -> int:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                row = await conn.fetchrow(
                    "select user_id, new_email, valid_until from pending_email_change where token = $1",
                    token,
                )
                user_id = row["user_id"]
                valid_until = row["valid_until"]
                if valid_until is None or valid_until < datetime.now(tz=timezone.utc):
                    raise PermissionError

                await conn.execute(
                    "delete from pending_email_change where user_id = $1", user_id
                )
                await conn.execute(
                    "update usr set email = $2 where id = $1", user_id, row["new_email"]
                )

                return user_id

    async def request_password_recovery(self, email: str):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                user_id = await conn.fetchval(
                    "select id from usr where email = $1", email
                )
                if not user_id:
                    raise PermissionError

                await conn.execute(
                    "insert into pending_password_recovery (user_id) values ($1)",
                    user_id,
                )

    async def confirm_password_recovery(self, token: str, new_password: str) -> int:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                row = await conn.fetchrow(
                    "select user_id, valid_until from pending_password_recovery where token = $1",
                    token,
                )
                user_id = row["user_id"]
                valid_until = row["valid_until"]
                if valid_until is None or valid_until < datetime.now(tz=timezone.utc):
                    raise PermissionError

                await conn.execute(
                    "delete from pending_password_recovery where user_id = $1", user_id
                )
                hashed_password = self._hash_password(password=new_password)
                await conn.execute(
                    "update usr set hashed_password = $2 where id = $1",
                    user_id,
                    hashed_password,
                )

                return user_id



import argparse
import asyncio
from pathlib import Path

from abrechnung.config import Config
from abrechnung.util import log_setup


def main():
    """
    main CLI entry point

    parses commands and jumps into the subprogram's entry point
    """
    cli = argparse.ArgumentParser()

    cli.add_argument(
        "-c",
        "--config-path",
        default="/etc/abrechnung/abrechnung.yaml",
        help="config file, default: %(default)s",
    )

    cli.add_argument(
        "-d", "--debug", action="store_true", help="enable asyncio debugging"
    )
    cli.add_argument(
        "-v", "--verbose", action="count", default=0, help="increase program verbosity"
    )
    cli.add_argument(
        "-q", "--quiet", action="count", default=0, help="decrease program verbosity"
    )

    subparsers = cli.add_subparsers()

    def add_subcommand(name, subcommand_class):
        subparser = subparsers.add_parser(name)
        subparser.set_defaults(subcommand_class=subcommand_class)
        subcommand_class.argparse_register(subparser)

    # add all the subcommands here
    from . import mailer

    add_subcommand("mailer", mailer.Mailer)

    from . import http

    add_subcommand("api", http.HTTPService)

    from . import database

    add_subcommand("db", database.CLI)

    from . import admin

    add_subcommand("admin", admin.Admin)

    args = vars(cli.parse_args())

    # set up log level
    log_setup(args["verbose"] - args["quiet"])

    # enable asyncio debugging
    # loop = asyncio.get_event_loop()
    # loop.set_debug(args["debug"])

    config_path = args.pop("config_path")
    config = Config.from_file(Path(config_path))

    try:
        subcommand_class = args.pop("subcommand_class")
    except KeyError:
        cli.error("no subcommand was given")
    subcommand_class.argparse_validate(args, cli.error)
    subcommand_object = subcommand_class(config=config, **args)
    asyncio.run(subcommand_object.run())


if __name__ == "__main__":
    main()



import jwt

from abrechnung.http.auth import token_for_user
from tests.http_tests import BaseHTTPAPITest


class AuthAPITest(BaseHTTPAPITest):
    async def _login(
        self,
        username: str,
        password: str,
        session_name: str = "dummy session",
        expected_status: int = 200,
    ):
        resp = await self.client.post(
            f"/api/v1/auth/login",
            json={
                "username": username,
                "password": password,
                "session_name": session_name,
            },
        )
        self.assertEqual(expected_status, resp.status)
        return await resp.json()

    async def _fetch_profile(self, token: str, expected_status: int = 200):
        resp = await self.client.get(
            f"/api/v1/profile", headers={"Authorization": f"Bearer {token}"}
        )
        self.assertEqual(expected_status, resp.status)
        if resp.status < 400:
            return await resp.json()

        return None

    async def test_register_user(self):
        resp = await self.client.post(
            f"/api/v1/auth/register",
            json={
                "username": "user",
                "email": "email@mail.com",
                "password": "password",
            },
        )
        self.assertEqual(200, resp.status)
        ret_data = await resp.json()
        self.assertIsNotNone(ret_data["user_id"])

        resp = await self.client.post(
            f"/api/v1/auth/register",
            json={
                "username": "user2",
                "email": "email@mail.com",
                "password": "password",
            },
        )
        self.assertEqual(400, resp.status)

    async def test_login_user(self):
        user_name = "user"
        email = "email@email.com"
        user_id = await self.user_service.register_user(
            username=user_name, email=email, password="password"
        )
        await self._login(user_name, "password", expected_status=400)

        # fetch the registration token from the database
        async with self.db_pool.acquire() as conn:
            token = await conn.fetchval(
                "select token from pending_registration where user_id = $1", user_id
            )

        resp = await self.client.post(
            f"/api/v1/auth/confirm_registration", json={"token": str(token)}
        )
        self.assertEqual(204, resp.status)

        # now we should be able to login and get a session token
        resp = await self._login(email, "password")
        self.assertIsNotNone(resp["user_id"])
        self.assertIsNotNone(resp["access_token"])
        self.assertIsNotNone(resp["session_token"])

        resp = await self._login("user", "password")

        token = resp["access_token"]
        # now check that we can actually fetch our profile with the token
        ret_data = await self._fetch_profile(token)
        self.assertEqual(user_id, ret_data["id"])
        self.assertEqual(user_name, ret_data["username"])
        self.assertEqual(email, ret_data["email"])
        # we logged in twice, once with our email, once with our username
        self.assertEqual(2, len(ret_data["sessions"]))

        # also check that a random token does not pass
        await self._fetch_profile("foolol", expected_status=401)

        invalid_token = jwt.encode(
            {"user_id": user_id}, key="very_secret_invalid_key", algorithm="HS256"
        )
        await self._fetch_profile(invalid_token, expected_status=401)

        # now check that we can logout and afterwards not fetch the profile anymore
        resp = await self.client.post(
            f"/api/v1/auth/logout", headers={"Authorization": f"Bearer {token}"}
        )
        self.assertEqual(204, resp.status)
        await self._fetch_profile(token, expected_status=401)

    async def test_change_password(self):
        user_id, password = await self._create_test_user("user", "user@email.stuff")
        _, session_id, _ = await self.user_service.login_user(
            "user", password=password, session_name="session1"
        )
        token = token_for_user(user_id, session_id, self.secret_key)

        headers = {"Authorization": f"Bearer {token}"}

        # check that we cannot change anything without providing the correct password
        resp = await self.client.post(
            f"/api/v1/profile/change_password",
            headers=headers,
            json={"old_password": "foobar", "new_password": "password3"},
        )
        self.assertEqual(400, resp.status)

        resp = await self.client.post(
            f"/api/v1/profile/change_password",
            headers=headers,
            json={"old_password": password, "new_password": "password2"},
        )
        self.assertEqual(204, resp.status)

        # check that we can login with the new password
        await self._login("user", "password2")

    async def test_change_email(self):
        username = "user1"
        old_email = "user@stusta.de"
        new_email = "new_email@stusta.de"
        user_id, password = await self._create_test_user(username, old_email)
        _, session_id, _ = await self.user_service.login_user(
            username=username, password=password, session_name="session1"
        )
        token = token_for_user(user_id, session_id, self.secret_key)

        headers = {"Authorization": f"Bearer {token}"}
        resp = await self.client.post(
            f"/api/v1/profile/change_email",
            headers=headers,
            json={"email": new_email, "password": password},
        )
        self.assertEqual(204, resp.status)

        resp = await self.client.post(
            f"/api/v1/profile/change_email",
            headers=headers,
            json={"email": new_email, "password": "asdf1234"},
        )
        self.assertEqual(400, resp.status)

        resp = await self.client.get(f"/api/v1/profile", headers=headers)
        self.assertEqual(200, resp.status)
        profile = await resp.json()
        # we still have the old email
        self.assertEqual(old_email, profile["email"])

        # confirm the email change
        # fetch the registration token from the database
        async with self.db_pool.acquire() as conn:
            token = await conn.fetchval(
                "select token from pending_email_change where user_id = $1", user_id
            )
        resp = await self.client.post(
            f"/api/v1/auth/confirm_email_change", json={"token": "foobar lol"}
        )
        self.assertEqual(400, resp.status)

        resp = await self.client.post(
            f"/api/v1/auth/confirm_email_change", json={"token": str(token)}
        )
        self.assertEqual(204, resp.status)

        # now we have the new email
        resp = await self.client.get(f"/api/v1/profile", headers=headers)
        self.assertEqual(200, resp.status)
        profile = await resp.json()
        # we still have the old email
        self.assertEqual(new_email, profile["email"])

    async def test_reset_password(self):
        user_email = "user@stusta.de"
        username = "user1"
        user_id, password = await self._create_test_user(username, user_email)
        resp = await self.client.post(
            f"/api/v1/auth/recover_password",
            json={"email": "fooo@stusta.de"},
        )
        self.assertEqual(403, resp.status)

        resp = await self.client.post(
            f"/api/v1/auth/recover_password",
            json={"email": user_email},
        )
        self.assertEqual(204, resp.status)

        # confirm the email change
        # fetch the registration token from the database
        async with self.db_pool.acquire() as conn:
            token = await conn.fetchval(
                "select token from pending_password_recovery where user_id = $1",
                user_id,
            )

        self.assertIsNotNone(token)

        resp = await self.client.post(
            f"/api/v1/auth/confirm_password_recovery",
            json={"token": "foobar", "new_password": "secret secret"},
        )
        self.assertEqual(400, resp.status)

        resp = await self.client.post(
            f"/api/v1/auth/confirm_password_recovery",
            json={"token": str(token), "new_password": "new_password"},
        )
        self.assertEqual(204, resp.status)

        # check that we can login with the new password
        await self._login(username, "new_password")

    async def test_sessions(self):
        user_email = "user@email.email"
        username = "user1"
        user_id, password = await self._create_test_user(username, user_email)

        resp = await self._login(username, password, session_name="session1")
        token = resp["access_token"]
        headers = {"Authorization": f"Bearer {token}"}

        profile = await self._fetch_profile(token)
        self.assertEqual(1, len(profile["sessions"]))
        self.assertEqual("session1", profile["sessions"][0]["name"])
        session_id = profile["sessions"][0]["id"]

        resp = await self.client.post(
            "/api/v1/auth/rename_session",
            json={"session_id": session_id, "name": "new_session_name"},
            headers=headers,
        )
        self.assertEqual(204, resp.status)

        profile = await self._fetch_profile(token)
        self.assertEqual(1, len(profile["sessions"]))
        self.assertEqual("new_session_name", profile["sessions"][0]["name"])
        session_id = profile["sessions"][0]["id"]

        resp = await self.client.post(
            "/api/v1/auth/delete_session",
            json={"session_id": session_id},
            headers=headers,
        )
        self.assertEqual(204, resp.status)
        await self._fetch_profile(token, expected_status=401)



import logging
import re
from pathlib import Path
from typing import Optional

import asyncpg
from asyncpg.pool import Pool

REVISION_DIR = Path(__file__).parent
logger = logging.getLogger(__name__)

REVISION_VERSION_RE = re.compile(r"^-- revision: (?P<version>\w+)$")
REVISION_REQUIRES_RE = re.compile(r"^-- requires: (?P<version>\w+)$")
REVISION_TABLE = "schema_revision"


class SchemaRevision:
    def __init__(
        self, file_name: Path, code: str, version: str, requires: Optional[str]
    ):
        self.file_name = file_name
        self.code = code
        self.version = version
        self.requires = requires

    async def apply(self, conn: asyncpg.Connection):
        logger.info(
            f"Applying revision {self.file_name.name} with version {self.version}"
        )
        if self.requires:
            version = await conn.fetchval(
                f"update {REVISION_TABLE} set version = $1 where version = $2 returning version",
                self.version,
                self.requires,
            )
            if version != self.version:
                raise ValueError(
                    f"Found other revision present than {self.requires} which was required"
                )
        else:
            n_table = await conn.fetchval(f"select count(*) from {REVISION_TABLE}")
            if n_table != 0:
                raise ValueError(
                    f"Could not apply revision {self.version} as there appears to be a revision present,"
                    f"none was expected"
                )
            await conn.execute(
                f"insert into {REVISION_TABLE} (version) values ($1)", self.version
            )

        # now we can actually apply the revision
        await conn.execute(self.code)

    @classmethod
    def revisions_from_dir(cls, file_path: Path) -> list["SchemaRevision"]:
        """
        returns an ordered list of revisions with their dependencies resolved
        """
        revisions = []
        for revision in sorted(REVISION_DIR.glob("*.sql")):
            revision_content = revision.read_text("utf-8")
            lines = revision_content.splitlines()
            if not len(lines) > 2:
                raise ValueError(f"Revision {revision} is invalid")

            if (version_match := REVISION_VERSION_RE.match(lines[0])) is None:
                raise ValueError(
                    f"Invalid version string in revision {revision}, "
                    f"should be of form '-- revision: <name>'"
                )
            if (requires_match := REVISION_REQUIRES_RE.match(lines[1])) is None:
                raise ValueError(
                    f"Invalid requires string in revision {revision}, "
                    f"should be of form '-- requires: <name>'"
                )

            version = version_match["version"]
            requires: Optional[str] = requires_match["version"]

            if requires == "null":
                requires = None

            revisions.append(
                cls(
                    revision,
                    revision_content,
                    version,
                    requires,
                )
            )

        if len(revisions) == 0:
            return revisions

        # now for the purpose of sorting the revisions according to their dependencies
        first_revision = next((x for x in revisions if x.requires is None), None)
        if first_revision is None:
            raise ValueError(f"Could not find a revision without any dependencies")

        # TODO: detect revision branches
        sorted_revisions = [first_revision]
        while len(sorted_revisions) < len(revisions):
            curr_revision = sorted_revisions[-1]
            next_revision = next(
                (x for x in revisions if x.requires == curr_revision.version), None
            )
            if next_revision is None:
                raise ValueError(
                    f"Could not find the successor to revision {curr_revision.version}"
                )
            sorted_revisions.append(next_revision)

        return sorted_revisions


async def reset_schema(db_pool: Pool):
    async with db_pool.acquire() as conn:
        async with conn.transaction():
            await conn.execute("drop schema public cascade")
            await conn.execute("create schema public")


async def apply_revisions(db_pool: Pool):
    revisions = SchemaRevision.revisions_from_dir(REVISION_DIR)

    async with db_pool.acquire() as conn:
        async with conn.transaction():
            await conn.execute(
                f"create table if not exists {REVISION_TABLE} ("
                "  version text not null primary key"
                ")"
            )

            curr_revision = await conn.fetchval(
                f"select version from {REVISION_TABLE} limit 1"
            )

            found = curr_revision is None
            for revision in revisions:
                if found:
                    await revision.apply(conn)

                if revision.version == curr_revision:
                    found = True

            if not found:
                raise ValueError(
                    f"Unknown revision {curr_revision} present in database"
                )



from typing import Any, Mapping, Optional

from marshmallow import Schema, fields


class GroupSchema(Schema):
    id = fields.Int()
    name = fields.Str()
    description = fields.Str()
    currency_symbol = fields.Str()
    terms = fields.Str()
    created_by = fields.Int()


class GroupPreviewSchema(Schema):
    id = fields.Int()
    name = fields.Str()
    description = fields.Str()
    currency_symbol = fields.Str()
    terms = fields.Str()
    created_at = fields.DateTime()
    invite_single_use = fields.Bool()
    invite_valid_until = fields.DateTime()
    invite_description = fields.Str()


class AccountSchema(Schema):
    id = fields.Int()
    group_id = fields.Int()
    type = fields.Str()
    name = fields.Str()
    description = fields.Str()
    deleted = fields.Bool()


class GroupInviteSchema(Schema):
    id = fields.Int()
    created_by = fields.Int()
    single_use = fields.Bool()
    valid_until = fields.DateTime()
    token = fields.Str()
    description = fields.Str()
    join_as_editor = fields.Bool()


class GroupLogSchema(Schema):
    id = fields.Int()
    type = fields.Str()
    message = fields.Str()
    user_id = fields.Int()
    logged_at = fields.DateTime()
    affected_user_id = fields.Int()


class SharesField(fields.Field):
    def _serialize(self, value: Mapping[int, float], attr: str, obj: Any, **kwargs):
        return {
            str(account_id): usage_value for account_id, usage_value in value.items()
        }

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ):
        raise NotImplementedError()


class TransactionPositionSchema(Schema):
    id = fields.Int()
    price = fields.Number()
    communist_shares = fields.Number()
    deleted = fields.Bool()
    name = fields.Str()
    usages = SharesField()


class TransactionDetailSchema(Schema):
    description = fields.Str()
    value = fields.Number()
    currency_symbol = fields.Str()
    currency_conversion_rate = fields.Number()
    billed_at = fields.Date()
    committed_at = fields.DateTime(required=False)
    creditor_shares = SharesField()
    debitor_shares = SharesField()
    deleted = fields.Bool()


class FileAttachmentSchema(Schema):
    id = fields.Int()
    filename = fields.Method("get_filename")
    blob_id = fields.Int(allow_none=True)
    deleted = fields.Bool()
    url = fields.Method("get_url")

    def get_filename(self, obj):
        return (
            obj.filename + ("." + obj.mime_type.split("/")[1]) if obj.mime_type else ""
        )

    def get_url(self, obj):
        return f"{obj.host_url}/v1/files/{obj.id}/{obj.blob_id}"


class TransactionSchema(Schema):
    id = fields.Int()
    type = fields.Str()
    is_wip = fields.Bool()
    last_changed = fields.DateTime()
    group_id = fields.Int()
    version = fields.Int()
    pending_details = fields.Nested(TransactionDetailSchema, allow_none=True)
    committed_details = fields.Nested(TransactionDetailSchema, allow_none=True)
    pending_positions = fields.List(
        fields.Nested(TransactionPositionSchema), dump_default=[]
    )
    committed_positions = fields.List(
        fields.Nested(TransactionPositionSchema), dump_default=[]
    )
    pending_files = fields.List(fields.Nested(FileAttachmentSchema), dump_default=[])
    committed_files = fields.List(fields.Nested(FileAttachmentSchema), dump_default=[])


class SessionSchema(Schema):
    id = fields.Int()
    name = fields.Str()
    valid_until = fields.DateTime()
    last_seen = fields.DateTime()


class UserSchema(Schema):
    id = fields.Int()
    username = fields.Str()
    email = fields.Email()
    sessions = fields.List(fields.Nested(SessionSchema))


class GroupMemberSchema(Schema):
    user_id = fields.Int()
    username = fields.Str()
    is_owner = fields.Bool()
    can_write = fields.Bool()
    description = fields.Str()
    joined_at = fields.DateTime()
    invited_by = fields.Int()



import argparse
import logging
from getpass import getpass

from abrechnung import subcommand
from abrechnung.application.users import UserService
from abrechnung.config import Config
from abrechnung.database import db_connect


class Admin(subcommand.SubCommand):
    def __init__(self, config: Config, **args):  # pylint: disable=super-init-not-called
        self.config = config
        self.logger = logging.getLogger(__name__)

        self.command = args.pop("command")
        self.args = args

    @staticmethod
    def argparse_register(subparser: argparse.ArgumentParser):
        subparsers = subparser.add_subparsers(required=True)

        create_user_parser = subparsers.add_parser("create-user")
        create_user_parser.set_defaults(command="create_user")
        create_user_parser.add_argument("-n", "--name", type=str, required=True)
        create_user_parser.add_argument("-e", "--email", type=str, required=True)
        create_user_parser.add_argument("--skip-email-check", action="store_true")

    async def handle_create_user_command(self):
        self.logger.info(
            f"Creating user with email: {self.args['email']} and username: {self.args['name']}"
        )
        password = getpass(prompt="Input initial password for user:")
        repeat_password = getpass(prompt="Repeat password:")
        if password != repeat_password:
            print("Passwords do not match!")
            return

        db_pool = await db_connect(
            username=self.config["database"]["user"],
            database=self.config["database"]["dbname"],
            host=self.config["database"]["host"],
            password=self.config["database"]["password"],
        )
        user_service = UserService(db_pool, self.config)
        user_service.enable_registration = True
        if self.args["skip_email_check"]:
            user_service.valid_email_domains = None
        await user_service.register_user(
            username=self.args["name"], email=self.args["email"], password=password
        )

    async def run(self):
        if self.command == "create_user":
            return await self.handle_create_user_command()



import contextlib
import os
import shutil
import tempfile

import asyncpg
from asyncpg.pool import Pool

from abrechnung import subcommand
from abrechnung import util
from abrechnung.config import Config
from . import revisions


async def db_connect(
    username: str, password: str, database: str, host: str, port: int = 5432
) -> Pool:
    """
    get a connection pool to the database
    """
    return await asyncpg.create_pool(
        user=username,
        password=password,
        database=database,
        host=host,
        port=port,
        max_size=100,
    )


class CLI(subcommand.SubCommand):
    """
    PSQL command.

    Runs the 'psql' tool with the appropriate parameters.
    """

    def __init__(self, config: Config, **args):
        self.config = config
        self.action = args["action"]

    @staticmethod
    def argparse_register(subparser):
        subparser.add_argument(
            "action", choices=["migrate", "attach", "rebuild", "clean"], nargs="?"
        )

    async def _clean(self, db_pool: Pool):
        """Do some garbage collection in the database"""
        async with db_pool.acquire() as conn:
            async with conn.transaction():
                # delete all dangling file contents
                # TODO: make this more future proof by using some kind of introspection to find all non-referenced
                # entries in the table and delete them
                await conn.execute(
                    "delete from blob where id not in ("
                    "   select fh.blob_id from file_history fh where fh.blob_id is not null"
                    ")"
                )

    async def _attach(self):
        with contextlib.ExitStack() as exitstack:
            env = dict(os.environ)
            env["PGDATABASE"] = self.config["database"]["dbname"]

            if self.config["database"]["user"] is None:
                if self.config["database"]["host"] is not None:
                    raise ValueError("database user is None, but host is set")
                if self.config["database"]["password"] is not None:
                    raise ValueError("database user is None, " "but password is set")
            else:

                def escape_colon(str):
                    return str.replace("\\", "\\\\").replace(":", "\\:")

                passfile = exitstack.enter_context(tempfile.NamedTemporaryFile("w"))
                os.chmod(passfile.name, 0o600)

                passfile.write(
                    ":".join(
                        [
                            escape_colon(self.config["database"]["host"]),
                            "*",
                            escape_colon(self.config["database"]["dbname"]),
                            escape_colon(self.config["database"]["user"]),
                            escape_colon(self.config["database"]["password"]),
                        ]
                    )
                )
                passfile.write("\n")
                passfile.flush()

                env["PGHOST"] = self.config["database"]["host"]
                env["PGUSER"] = self.config["database"]["user"]
                env["PGPASSFILE"] = passfile.name

            command = ["psql", "--variable", "ON_ERROR_STOP=1"]
            if shutil.which("pgcli") is not None:
                # if pgcli is installed, use that instead!
                command = ["pgcli"]

            cwd = os.path.join(os.path.dirname(__file__), "revisions")
            ret = await util.run_as_fg_process(command, env=env, cwd=cwd)

            if ret != 0:
                print(util.format_error("psql failed"))
            return ret

    async def run(self):
        """
        CLI entry point
        """
        if self.action == "attach":
            return await self._attach()

        db_pool = await db_connect(
            username=self.config["database"]["user"],
            database=self.config["database"]["dbname"],
            host=self.config["database"]["host"],
            password=self.config["database"]["password"],
        )
        if self.action == "migrate":
            await revisions.apply_revisions(db_pool=db_pool)
        elif self.action == "rebuild":
            await revisions.reset_schema(db_pool=db_pool)
            await revisions.apply_revisions(db_pool=db_pool)
        elif self.action == "clean":
            await self._clean(db_pool=db_pool)



import asyncio
import email.message
import email.utils
import itertools
import logging
import smtplib

import asyncpg

from . import subcommand
from .config import Config


class Mailer(subcommand.SubCommand):
    def __init__(self, config: Config, **args):  # pylint: disable=super-init-not-called
        del args  # unused

        self.config = config
        self.events: asyncio.Queue = asyncio.Queue()
        self.psql = None
        self.mailer = None
        self.logger = logging.getLogger(__name__)

        self.event_handlers = {
            (
                "mailer",
                "pending_registration",
            ): self.on_pending_registration_notification,
            (
                "mailer",
                "pending_password_recovery",
            ): self.on_user_password_recovery_notification,
            ("mailer", "pending_email_change"): self.on_user_email_update_notification,
        }

    async def run(self):
        # just try to connect to the mailing server once
        _ = self.get_mailer_instance()

        self.psql = await asyncpg.connect(
            user=self.config["database"]["user"],
            password=self.config["database"]["password"],
            host=self.config["database"]["host"],
            database=self.config["database"]["dbname"],
            port=self.config["database"].get("port", 5432),
        )
        self.psql.add_termination_listener(self.terminate_callback)
        self.psql.add_log_listener(self.log_callback)
        await self.psql.add_listener("mailer", self.notification_callback)

        # run all of the events manually once
        for handler in self.event_handlers.values():
            await handler()

        # handle events
        while True:
            event = await self.events.get()
            if event is StopIteration:
                break
            handler = self.event_handlers.get(event)
            if handler is None:
                self.logger.info(f"unhandled event {event!r}")
            else:
                await handler()

        await self.psql.remove_listener("mailer", self.notification_callback)
        await self.psql.close()

    def get_mailer_instance(self):
        mode = self.config["email"].get("mode")

        if mode == "local":
            mail_sender_class = smtplib.LMTP
        elif mode == "smtp-ssl":
            mail_sender_class = smtplib.SMTP_SSL
        else:
            mail_sender_class = smtplib.SMTP

        mailer = mail_sender_class(
            host=self.config["email"]["host"],
            port=self.config["email"]["port"],
        )
        if mode == "smtp-starttls":
            mailer.starttls()

        if "auth" in self.config["email"]:
            mailer.login(
                user=self.config["email"]["auth"]["username"],
                password=self.config["email"]["auth"]["password"],
            )
        return mailer

    def notification_callback(
        self, connection: asyncpg.Connection, pid: int, channel: str, payload: str
    ):
        """runs whenever we get a psql notification"""
        assert connection is self.psql
        del pid  # unused
        self.events.put_nowait((channel, payload))

    def terminate_callback(self, connection: asyncpg.Connection):
        """runs when the psql connection is closed"""
        assert connection is self.psql
        self.logger.info("psql connection closed")
        # proper way of clearing asyncio queue
        for _ in range(self.events.qsize()):
            self.events.get_nowait()
            self.events.task_done()
        self.events.put_nowait(StopIteration)

    async def log_callback(self, connection: asyncpg.Connection, message: str):
        """runs when psql sends a log message"""
        assert connection is self.psql
        self.logger.info(f"psql log message: {message}")

    def send_email(
        self, *text_lines: str, subject: str, dest_address: str, dest_name: str
    ):
        self.logger.info(f"sending email to {dest_address}, subject: {subject}")

        # we do this to not have one long hanging open connection with the mail server
        mailer = self.get_mailer_instance()

        from_addr = self.config["email"]["address"]
        msg = email.message.EmailMessage()
        msg.set_content(
            "\n".join(
                itertools.chain(
                    self.greeting_lines(dest_name),
                    text_lines,
                    self.closing_lines(),
                )
            )
        )
        msg["Subject"] = f"[{self.config['service']['name']}] {subject}"
        msg["To"] = dest_address
        msg["From"] = from_addr
        msg["Date"] = email.utils.localtime()
        msg["Message-ID"] = email.utils.make_msgid(domain=from_addr.split("@")[-1])
        mailer.send_message(msg)  # type: ignore

    def greeting_lines(self, name: str):
        return f"Beloved {name},", ""

    def closing_lines(self):
        return "", "Thoughtfully yours", "", f"    {self.config['service']['name']}"

    async def on_pending_registration_notification(self):
        unsent_mails = await self.psql.fetch(
            "select usr.id, usr.email, usr.username, pr.token, pr.valid_until "
            "from pending_registration pr join usr on usr.id = pr.user_id "
            "where pr.mail_next_attempt is not null and pr.mail_next_attempt < NOW() and pr.valid_until > NOW()"
        )

        if not unsent_mails:
            self.logger.info("no pending_registration mails are pending")

        for row in unsent_mails:
            try:
                self.send_email(
                    "it looks like you are attempting to create a user account.",
                    "",
                    "To complete your registration, visit",
                    "",
                    f"{self.config['service']['url']}/confirm-registration/{row['token']}",
                    "",
                    f"Your request will time out {row['valid_until']}.",
                    "If you do not want to create a user account, just ignore this email.",
                    subject="Confirm user account",
                    dest_address=row["email"],
                    dest_name=row["username"],
                )

                await self.psql.execute(
                    "update pending_registration "
                    "set mail_next_attempt = null "
                    "where token = $1",
                    row["token"],
                )
            except smtplib.SMTPException as e:
                self.logger.warning(
                    f"Failed to send email to user {row['username']} with email {row['email']}: {e}"
                )

    async def on_user_password_recovery_notification(self):
        unsent_mails = await self.psql.fetch(
            "select usr.id, usr.username, usr.email, ppr.token, ppr.valid_until "
            "from pending_password_recovery ppr join usr on usr.id = ppr.user_id "
            "where ppr.mail_next_attempt is not null and ppr.mail_next_attempt < NOW() and ppr.valid_until > NOW()"
        )

        if not unsent_mails:
            self.logger.info("no user_password_recovery mails are pending")

        for row in unsent_mails:
            try:
                self.send_email(
                    "it looks like you forgot your password; how embarrasing.",
                    "",
                    "To set a new one, visit",
                    "",
                    f"{self.config['service']['url']}/confirm-password-recovery/{row['token']}",
                    "",
                    f"Your request will time out {row['valid_until']}.",
                    "If you do not want to reset your password, just ignore this email.",
                    subject="Reset password",
                    dest_address=row["email"],
                    dest_name=row["username"],
                )

                await self.psql.execute(
                    "update pending_password_recovery "
                    "set mail_next_attempt = null "
                    "where token = $1",
                    row["token"],
                )
            except smtplib.SMTPException as e:
                self.logger.warning(
                    f"Failed to send email to user {row['username']} with email {row['email']}: {e}"
                )

    async def on_user_email_update_notification(self):
        unsent_mails = await self.psql.fetch(
            "select usr.id, usr.username, usr.email as old_email, pec.new_email as new_email, pec.token, "
            "   pec.valid_until "
            "from pending_email_change pec join usr on usr.id = pec.user_id "
            "where pec.mail_next_attempt is not null and pec.mail_next_attempt < NOW() and pec.valid_until > NOW()"
        )

        if not unsent_mails:
            self.logger.info("no user_email_update mails are pending")

        for row in unsent_mails:
            try:
                self.send_email(
                    "you want to change your email address",
                    "",
                    f"Your current email is: {row['old_email']}",
                    f"You want to change it to: {row['new_email']}",
                    "",
                    "To confirm, see the mail that was sent to the new address.",
                    "",
                    f"Your request will time out {row['valid_until']}.",
                    "If you do not want to change your email, just ignore this email.",
                    subject="Change email",
                    dest_address=row["old_email"],
                    dest_name=row["username"],
                )

                self.send_email(
                    "you want to change your email address",
                    "",
                    f"Your current email is: {row['old_email']}",
                    f"You want to change it to: {row['new_email']}",
                    "",
                    "To confirm, visit",
                    "",
                    f"{self.config['service']['url']}/confirm-email-change/{row['token']}",
                    "",
                    f"Your request will time out {row['valid_until']}.",
                    "If you do not want to change your email, just ignore this email.",
                    subject="Change email",
                    dest_address=row["new_email"],
                    dest_name=row["username"],
                )

                await self.psql.execute(
                    "update pending_email_change "
                    "set mail_next_attempt = null "
                    "where token = $1",
                    row["token"],
                )
            except smtplib.SMTPException as e:
                self.logger.warning(
                    f"Failed to send email to user {row['username']} with email {row['email']}: {e}"
                )



from dataclasses import dataclass
from enum import Enum


class AccountType(Enum):
    personal = "personal"


@dataclass
class Account:
    id: int
    group_id: int
    type: str
    name: str
    description: str
    priority: int
    # created_by: int
    deleted: bool



from pathlib import Path

import yaml
from marshmallow import Schema, fields

CONFIG_SCHEMA = Schema.from_dict(
    name="ConfigSchema",
    fields={
        "service": fields.Nested(
            Schema.from_dict(
                {"url": fields.Str(), "name": fields.Str(), "api_url": fields.Str()}
            )
        ),
        "database": fields.Nested(
            Schema.from_dict(
                {
                    "host": fields.Str(),
                    "port": fields.Int(missing=5432, required=False),
                    "user": fields.Str(),
                    "dbname": fields.Str(),
                    "password": fields.Str(),
                }
            )
        ),
        "api": fields.Nested(
            Schema.from_dict(
                {
                    "secret_key": fields.Str(),
                    "host": fields.Str(),
                    "port": fields.Int(),
                    "id": fields.Str(),
                    "max_uploadable_file_size": fields.Int(
                        missing=1024, required=False
                    ),  # in KB
                    "enable_cors": fields.Bool(required=False, missing=True),
                    "enable_registration": fields.Bool(required=False, missing=False),
                    "valid_email_domains": fields.List(
                        fields.Str(), required=False, missing=None
                    ),
                }
            )
        ),
        "email": fields.Nested(
            Schema.from_dict(
                {
                    "address": fields.Str(),
                    "host": fields.Str(),
                    "port": fields.Int(),
                    "mode": fields.Str(required=False),
                    "auth": fields.Nested(
                        Schema.from_dict(
                            {"username": fields.Str(), "password": fields.Str()},
                            name="MailAuthSchema",
                        ),
                        required=False,
                    ),
                }
            )
        ),
    },
)


class Config:
    def __init__(self, cfg: dict):
        self._cfg = cfg

    def get(self, key, default=None):
        return self._cfg.get(key, default)

    def __getitem__(self, key):
        return self._cfg.__getitem__(key)

    @classmethod
    def from_dict(cls, cfg: dict) -> "Config":
        cfg = CONFIG_SCHEMA().load(cfg)
        return cls(cfg)

    @classmethod
    def from_file(cls, file_path: Path) -> "Config":
        with file_path.open("r") as f:
            cfg = yaml.safe_load(f)
            cfg = CONFIG_SCHEMA().load(cfg)
            return cls(cfg)



import os
from datetime import datetime

from abrechnung.application.accounts import AccountService
from abrechnung.application.groups import GroupService
from abrechnung.application.transactions import TransactionService
from abrechnung.application.users import UserService
from abrechnung.config import Config
from tests import AsyncTestCase


class TransactionLogicTest(AsyncTestCase):
    async def setUpAsync(self) -> None:
        await super().setUpAsync()
        self.secret_key = "asdf1234"
        self.test_config = Config.from_dict(
            {
                "api": {"secret_key": self.secret_key, "max_uploadable_file_size": 512},
                "service": {"api_url": "http://localhost/api"},
            }
        )
        self.group_service = GroupService(self.db_pool, config=self.test_config)
        self.account_service = AccountService(self.db_pool, config=self.test_config)
        self.user_service = UserService(self.db_pool, config=self.test_config)
        self.transaction_service = TransactionService(
            self.db_pool, config=self.test_config
        )

        self.user_id, _ = await self._create_test_user("test", "test@test.test")

    async def _create_accounts(self, group_id: int, n_accounts: int) -> list[int]:
        account_ids = []
        for i in range(n_accounts):
            acc_id = await self.account_service.create_account(
                user_id=self.user_id,
                group_id=group_id,
                type="personal",
                name=f"account{i}",
                description="",
            )
            account_ids.append(acc_id)

        return account_ids

    async def test_file_upload(self):
        group_id = await self.group_service.create_group(
            user_id=self.user_id,
            name="test group",
            description="",
            currency_symbol="€",
            terms="",
        )
        account1_id, account2_id = await self._create_accounts(group_id, 2)
        transaction_id = await self.transaction_service.create_transaction(
            user_id=self.user_id,
            group_id=group_id,
            type="purchase",
            description="foo",
            billed_at=datetime.now().date(),
            currency_symbol="€",
            currency_conversion_rate=1.0,
            value=33,
        )
        await self.transaction_service.add_or_change_debitor_share(
            user_id=self.user_id,
            transaction_id=transaction_id,
            account_id=account1_id,
            value=1.0,
        )
        await self.transaction_service.add_or_change_creditor_share(
            user_id=self.user_id,
            transaction_id=transaction_id,
            account_id=account2_id,
            value=1.0,
        )
        with open(
            os.path.join(os.path.dirname(__file__), "assets", "test_image.jpg"), "rb"
        ) as test_image:
            content = test_image.read()
            file_size = len(content)
            file_id = await self.transaction_service.upload_file(
                user_id=self.user_id,
                transaction_id=transaction_id,
                filename="test file",
                mime_type="image/jpeg",
                content=content,
            )
        transaction = await self.transaction_service.get_transaction(
            user_id=self.user_id, transaction_id=transaction_id
        )
        self.assertIsNotNone(transaction.pending_files)
        self.assertIsNone(transaction.committed_files)
        self.assertEqual(1, len(transaction.pending_files))
        self.assertEqual(file_id, transaction.pending_files[0].id)
        await self.transaction_service.commit_transaction(
            user_id=self.user_id, transaction_id=transaction_id
        )

        transaction = await self.transaction_service.get_transaction(
            user_id=self.user_id, transaction_id=transaction_id
        )
        self.assertIsNone(transaction.pending_files)
        self.assertIsNotNone(transaction.committed_files)
        self.assertEqual(1, len(transaction.committed_files))
        self.assertEqual(file_id, transaction.committed_files[0].id)
        self.assertIsNotNone(transaction.committed_files[0].blob_id)

        (mime_type, _,) = await self.transaction_service.read_file_contents(
            user_id=self.user_id,
            file_id=file_id,
            blob_id=transaction.committed_files[0].blob_id,
        )
        self.assertEqual("image/jpeg", mime_type)
        self.assertEqual(file_size, len(content))

        # now delete it
        await self.transaction_service.delete_file(
            user_id=self.user_id, file_id=file_id
        )
        transaction = await self.transaction_service.get_transaction(
            user_id=self.user_id, transaction_id=transaction_id
        )
        self.assertIsNotNone(transaction.pending_files)
        self.assertIsNotNone(transaction.committed_files)
        self.assertEqual(1, len(transaction.pending_files))
        self.assertEqual(file_id, transaction.pending_files[0].id)
        self.assertTrue(transaction.pending_files[0].deleted)
        await self.transaction_service.commit_transaction(
            user_id=self.user_id, transaction_id=transaction_id
        )
        transaction = await self.transaction_service.get_transaction(
            user_id=self.user_id, transaction_id=transaction_id
        )
        self.assertIsNone(transaction.pending_files)
        self.assertIsNotNone(transaction.committed_files)
        self.assertEqual(1, len(transaction.committed_files))
        self.assertEqual(file_id, transaction.committed_files[0].id)
        self.assertIsNone(transaction.committed_files[0].blob_id)
        self.assertTrue(transaction.committed_files[0].deleted)



from datetime import date, datetime

from tests.http_tests import HTTPAPITest


class TransactionAPITest(HTTPAPITest):
    async def _create_group_with_transaction(
        self, transaction_type: str
    ) -> tuple[int, int]:
        group_id = await self.group_service.create_group(
            user_id=self.test_user_id,
            name="name",
            description="description",
            currency_symbol="€",
            terms="terms",
        )
        transaction_id = await self.transaction_service.create_transaction(
            user_id=self.test_user_id,
            group_id=group_id,
            type=transaction_type,
            description="description123",
            currency_symbol="€",
            currency_conversion_rate=1.22,
            billed_at=date.today(),
            value=122.22,
        )
        return group_id, transaction_id

    async def _create_account(self, group_id: int, name: str) -> int:
        return await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name=name,
            description=f"account {name} description",
        )

    async def _fetch_transaction(
        self, transaction_id: int, expected_status: int = 200
    ) -> dict:
        resp = await self._get(f"/api/v1/transactions/{transaction_id}")
        self.assertEqual(expected_status, resp.status)
        ret_data = await resp.json()
        self.assertEqual(transaction_id, ret_data["id"])
        return ret_data

    async def _update_transaction(
        self,
        transaction_id: int,
        value: float,
        description: str,
        billed_at: date,
        currency_symbol: str,
        currency_conversion_rate: float,
        expected_status: int = 200,
    ) -> dict:
        resp = await self._post(
            f"/api/v1/transactions/{transaction_id}",
            json={
                "value": value,
                "description": description,
                "currency_symbol": currency_symbol,
                "currency_conversion_rate": currency_conversion_rate,
                "billed_at": billed_at.isoformat(),
            },
        )
        self.assertEqual(expected_status, resp.status)
        ret_data = await resp.json()
        return ret_data

    async def _commit_transaction(
        self, transaction_id: int, expected_status: int = 200
    ) -> None:
        resp = await self._post(f"/api/v1/transactions/{transaction_id}/commit")
        self.assertEqual(expected_status, resp.status)

    async def _create_change(
        self, transaction_id: int, expected_status: int = 200
    ) -> None:
        resp = await self._post(f"/api/v1/transactions/{transaction_id}/new_change")
        self.assertEqual(expected_status, resp.status)

    async def _delete_transaction(
        self, transaction_id: int, expected_status: int = 200
    ) -> None:
        resp = await self._delete(f"/api/v1/transactions/{transaction_id}")
        self.assertEqual(expected_status, resp.status)

    async def _discard_transaction_change(
        self, transaction_id: int, expected_status: int = 200
    ) -> None:
        resp = await self._post(f"/api/v1/transactions/{transaction_id}/discard")
        self.assertEqual(expected_status, resp.status)

    async def _post_creditor_share(
        self,
        transaction_id: int,
        account_id: int,
        value: float,
        expected_status: int = 200,
    ) -> None:
        resp = await self._post(
            f"/api/v1/transactions/{transaction_id}/creditor_shares",
            json={"account_id": account_id, "value": value},
        )
        self.assertEqual(expected_status, resp.status)

    async def _switch_creditor_share(
        self,
        transaction_id: int,
        account_id: int,
        value: float,
        expected_status: int = 200,
    ) -> None:
        resp = await self._post(
            f"/api/v1/transactions/{transaction_id}/creditor_shares/switch",
            json={"account_id": account_id, "value": value},
        )
        self.assertEqual(expected_status, resp.status)

    async def _delete_creditor_share(
        self,
        transaction_id: int,
        account_id: int,
        expected_status: int = 200,
    ) -> None:
        resp = await self._delete(
            f"/api/v1/transactions/{transaction_id}/creditor_shares",
            json={"account_id": account_id},
        )
        self.assertEqual(expected_status, resp.status)

    async def _post_debitor_share(
        self,
        transaction_id: int,
        account_id: int,
        value: float,
        expected_status: int = 200,
    ) -> None:
        resp = await self._post(
            f"/api/v1/transactions/{transaction_id}/debitor_shares",
            json={"account_id": account_id, "value": value},
        )
        self.assertEqual(expected_status, resp.status)

    async def _switch_debitor_share(
        self,
        transaction_id: int,
        account_id: int,
        value: float,
        expected_status: int = 200,
    ) -> None:
        resp = await self._post(
            f"/api/v1/transactions/{transaction_id}/debitor_shares/switch",
            json={"account_id": account_id, "value": value},
        )
        self.assertEqual(expected_status, resp.status)

    async def _delete_debitor_share(
        self,
        transaction_id: int,
        account_id: int,
        expected_status: int = 200,
    ) -> None:
        resp = await self._delete(
            f"/api/v1/transactions/{transaction_id}/debitor_shares",
            json={"account_id": account_id},
        )
        self.assertEqual(expected_status, resp.status)

    async def test_create_transaction(self):
        group_id = await self.group_service.create_group(
            user_id=self.test_user_id,
            name="name",
            description="description",
            currency_symbol="€",
            terms="terms",
        )
        resp = await self._post(
            f"/api/v1/groups/{group_id}/transactions",
            json={
                "description": "description",
                "type": "purchase",
                "currency_symbol": "€",
                "value": 123.22,
                "billed_at": date.today().isoformat(),
                "currency_conversion_rate": 1.33,
            },
        )
        self.assertEqual(200, resp.status)
        ret_data = await resp.json()
        self.assertIsNotNone(ret_data["id"])

    async def test_list_transactions(self):
        group_id = await self.group_service.create_group(
            user_id=self.test_user_id,
            name="name",
            description="description",
            currency_symbol="€",
            terms="terms",
        )
        transaction1_id = await self.transaction_service.create_transaction(
            user_id=self.test_user_id,
            group_id=group_id,
            type="purchase",
            description="description123",
            currency_symbol="€",
            billed_at=date.today(),
            currency_conversion_rate=1.22,
            value=122.22,
        )
        transaction2_id = await self.transaction_service.create_transaction(
            user_id=self.test_user_id,
            group_id=group_id,
            type="purchase",
            description="description123",
            currency_symbol="€",
            billed_at=date.today(),
            currency_conversion_rate=1.22,
            value=122.22,
        )
        resp = await self._get(f"/api/v1/groups/{group_id}/transactions")
        self.assertEqual(200, resp.status)
        ret_data = await resp.json()
        self.assertEqual(2, len(ret_data))
        self.assertEqual(
            {transaction1_id, transaction2_id},
            set([e["id"] for e in ret_data]),
        )

        transaction3_id = await self.transaction_service.create_transaction(
            user_id=self.test_user_id,
            group_id=group_id,
            type="purchase",
            description="foobar",
            currency_symbol="€",
            billed_at=date.today(),
            currency_conversion_rate=1,
            value=100,
        )
        account_id = await self._create_account(group_id=group_id, name="account1")
        await self._post_creditor_share(
            transaction_id=transaction3_id, account_id=account_id, value=1
        )
        await self._post_debitor_share(
            transaction_id=transaction3_id, account_id=account_id, value=1
        )

        # check that the list endpoint without parameters returns all objects
        await self.transaction_service.commit_transaction(
            user_id=self.test_user_id, transaction_id=transaction3_id
        )
        resp = await self._get(f"/api/v1/groups/{group_id}/transactions")
        self.assertEqual(200, resp.status)
        ret_data = await resp.json()
        self.assertEqual(3, len(ret_data))

        # TODO: test this aspect more thoroughly
        resp = await self._get(
            f"/api/v1/groups/{group_id}/transactions?min_last_changed={datetime.now().isoformat()}"
        )
        self.assertEqual(200, resp.status)
        ret_data = await resp.json()
        self.assertEqual(2, len(ret_data))
        self.assertNotIn(transaction3_id, [t["id"] for t in ret_data])

        resp = await self._get(
            f"/api/v1/groups/{group_id}/transactions?min_last_changed={datetime.now().isoformat()}&transaction_ids={transaction3_id}"
        )
        self.assertEqual(200, resp.status)
        ret_data = await resp.json()
        self.assertEqual(3, len(ret_data))

    async def test_get_transaction(self):
        group_id, transaction_id = await self._create_group_with_transaction("transfer")
        t = await self._fetch_transaction(transaction_id)
        self.assertIsNotNone(t["pending_details"])
        self.assertEqual(transaction_id, t["id"])

        resp = await self._get(f"/api/v1/transactions/asdf1234")
        self.assertEqual(404, resp.status)

        resp = await self._get(f"/api/v1/transactions/1332")
        self.assertEqual(404, resp.status)

    async def test_update_transaction(self):
        group_id, transaction_id = await self._create_group_with_transaction("transfer")
        await self._update_transaction(
            transaction_id, 200.0, "some description", date.today(), "$", 2.0
        )
        account1_id = await self._create_account(group_id, "account1")
        account2_id = await self._create_account(group_id, "account2")

        t = await self._fetch_transaction(transaction_id)
        self.assertEqual(200.0, t["pending_details"]["value"])
        self.assertEqual(
            "some description",
            t["pending_details"]["description"],
        )
        self.assertEqual("$", t["pending_details"]["currency_symbol"])
        self.assertEqual(
            2.0,
            t["pending_details"]["currency_conversion_rate"],
        )

        await self._post_debitor_share(transaction_id, account1_id, 1.0)
        await self._post_creditor_share(transaction_id, account2_id, 1.0)

        await self._commit_transaction(transaction_id)

        t = await self._fetch_transaction(transaction_id)
        self.assertIsNone(t["pending_details"])

        await self._update_transaction(
            transaction_id, 100.0, "foobar", date.today(), "€", 1.0
        )
        t = await self._fetch_transaction(transaction_id)
        self.assertEqual(100.0, t["pending_details"]["value"])
        self.assertEqual(
            "foobar",
            t["pending_details"]["description"],
        )
        self.assertEqual("€", t["pending_details"]["currency_symbol"])
        self.assertEqual(
            1.0,
            t["pending_details"]["currency_conversion_rate"],
        )

        await self._commit_transaction(transaction_id)
        t = await self._fetch_transaction(transaction_id)
        self.assertIsNone(t["pending_details"])

        await self._create_change(transaction_id)
        t = await self._fetch_transaction(transaction_id)
        self.assertTrue(t["is_wip"])
        await self._update_transaction(
            transaction_id, 200.0, "foofoo", date.today(), "$", 2.0
        )

        t = await self._fetch_transaction(transaction_id)
        self.assertEqual(200.0, t["pending_details"]["value"])
        self.assertEqual(
            "foofoo",
            t["pending_details"]["description"],
        )
        self.assertEqual("$", t["pending_details"]["currency_symbol"])
        self.assertEqual(
            2.0,
            t["pending_details"]["currency_conversion_rate"],
        )
        await self._commit_transaction(transaction_id)
        t = await self._fetch_transaction(transaction_id)
        self.assertEqual(200.0, t["committed_details"]["value"])
        self.assertEqual(
            "foofoo",
            t["committed_details"]["description"],
        )
        self.assertEqual("$", t["committed_details"]["currency_symbol"])
        self.assertEqual(
            2.0,
            t["committed_details"]["currency_conversion_rate"],
        )

    async def test_commit_transaction(self):
        group_id, transaction_id = await self._create_group_with_transaction("purchase")
        account1_id = await self._create_account(group_id, "account1")
        account2_id = await self._create_account(group_id, "account2")

        # we should not be able to commit this transaction as we do not have creditor or debitor shares
        await self._commit_transaction(transaction_id, expected_status=400)

        # create a creditor share and try to commit it, should not work as we do not have a debitor share
        await self._post_creditor_share(transaction_id, account1_id, 1.0)
        await self._commit_transaction(transaction_id, expected_status=400)

        await self._post_debitor_share(transaction_id, account2_id, 1.0)

        # now we should be able to commit
        await self._commit_transaction(transaction_id, expected_status=200)

        t = await self._fetch_transaction(transaction_id)
        self.assertIsNone(t["pending_details"])
        self.assertFalse(t["is_wip"])

        # test that we cannot commit without having pending changes
        await self._commit_transaction(transaction_id, expected_status=400)

        # create a second debitor share
        await self._post_debitor_share(transaction_id, account1_id, 1.0)

        # check that we have another pending change
        t = await self._fetch_transaction(transaction_id)
        self.assertIsNotNone(t["pending_details"])
        self.assertTrue(t["is_wip"])

        # check that we can commit this
        await self._commit_transaction(transaction_id)
        t = await self._fetch_transaction(transaction_id)
        self.assertIsNone(t["pending_details"])
        self.assertFalse(t["is_wip"])

        # try another edit and discard that
        await self._delete_debitor_share(transaction_id, account1_id)
        await self._discard_transaction_change(transaction_id)
        t = await self._fetch_transaction(transaction_id)
        self.assertIsNone(t["pending_details"])
        self.assertFalse(t["is_wip"])

        # try delete the transaction
        await self._delete_transaction(transaction_id)
        t = await self._fetch_transaction(transaction_id)
        self.assertTrue(t["committed_details"]["deleted"])

    async def test_discard_newly_created_transaction(self):
        group_id, transaction_id = await self._create_group_with_transaction("purchase")
        account1_id = await self._create_account(group_id, "account1")

        # one share for the sake of having it
        await self._post_debitor_share(transaction_id, account1_id, 1.0)

        # we should not be able to discard this transaction as it does not have any committed changes
        await self._discard_transaction_change(transaction_id, expected_status=400)

        await self._delete_transaction(transaction_id)

        t = await self._fetch_transaction(transaction_id)
        self.assertTrue(t["committed_details"]["deleted"])
        self.assertIsNone(t["pending_details"])
        self.assertFalse(t["is_wip"])

    async def test_creditor_shares(self):
        group_id, transaction_id = await self._create_group_with_transaction("purchase")
        account1_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account1",
            description="description",
        )
        account2_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account2",
            description="description",
        )

        await self._post_creditor_share(transaction_id, account1_id, 1.0)

        # test that we cannot add a second creditor share to a "purchase" type transaction
        await self._post_creditor_share(
            transaction_id, account2_id, 1.0, expected_status=400
        )

        # test that we can change the value of existing shares
        await self._post_creditor_share(transaction_id, account1_id, 2.0)

        # check the state
        t = await self._fetch_transaction(transaction_id)
        self.assertIn(
            str(account1_id),
            t["pending_details"]["creditor_shares"],
        )
        self.assertEqual(
            2.0,
            t["pending_details"]["creditor_shares"][str(account1_id)],
        )

        # delete one share
        await self._delete_creditor_share(transaction_id, account1_id)
        t = await self._fetch_transaction(transaction_id)
        self.assertNotIn(
            str(account1_id),
            t["pending_details"]["creditor_shares"],
        )

        # check that we cannot delete non existing shares
        await self._delete_creditor_share(
            transaction_id, account1_id, expected_status=404
        )

        # check that we can perform a switch creditor share on the "purchase" type transaction
        await self._switch_creditor_share(transaction_id, account1_id, 1.0)
        await self._switch_creditor_share(transaction_id, account2_id, 2.0)

        t = await self._fetch_transaction(transaction_id)
        self.assertEqual(
            t["pending_details"]["creditor_shares"][str(account2_id)],
            2.0,
        )

    async def test_debitor_shares_purchase(self):
        group_id, transaction_id = await self._create_group_with_transaction("purchase")
        account1_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account1",
            description="description",
        )
        account2_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account2",
            description="description",
        )

        await self._post_debitor_share(transaction_id, account1_id, 1.0)
        await self._post_debitor_share(transaction_id, account2_id, 1.0)

        # test that we can change the value of existing shares
        await self._post_debitor_share(transaction_id, account1_id, 2.0)

        # check the state
        t = await self._fetch_transaction(transaction_id)
        self.assertIn(
            str(account1_id),
            t["pending_details"]["debitor_shares"],
        )
        self.assertIn(
            str(account2_id),
            t["pending_details"]["debitor_shares"],
        )
        self.assertEqual(
            t["pending_details"]["debitor_shares"][str(account1_id)],
            2.0,
        )
        self.assertEqual(
            t["pending_details"]["debitor_shares"][str(account2_id)],
            1.0,
        )

        # delete one share
        await self._delete_debitor_share(transaction_id, account1_id)
        t = await self._fetch_transaction(transaction_id)
        self.assertNotIn(
            str(account1_id),
            t["pending_details"]["debitor_shares"],
        )

        # check that we cannot delete non existing shares
        await self._delete_debitor_share(
            transaction_id, account1_id, expected_status=404
        )

        # check that we cannot perform a switch debitor share on the "purchase" type transaction
        await self._switch_debitor_share(
            transaction_id, account1_id, 1.0, expected_status=400
        )

    async def test_debitor_shares_transfer(self):
        group_id, transaction_id = await self._create_group_with_transaction("transfer")
        account_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account1",
            description="description",
        )

        await self._post_debitor_share(transaction_id, account_id, 1.0)
        # test that we can change the value of existing shares
        await self._post_debitor_share(transaction_id, account_id, 2.0)

        # check the state
        t = await self._fetch_transaction(transaction_id)
        self.assertIn(
            str(account_id),
            t["pending_details"]["debitor_shares"],
        )
        self.assertEqual(
            t["pending_details"]["debitor_shares"][str(account_id)],
            2.0,
        )

        # delete one share
        await self._delete_debitor_share(transaction_id, account_id)
        t = await self._fetch_transaction(transaction_id)
        self.assertNotIn(
            str(account_id),
            t["pending_details"]["debitor_shares"],
        )

        # check that we cannot delete non existing shares
        await self._delete_debitor_share(
            transaction_id, account_id, expected_status=404
        )

        await self._switch_debitor_share(transaction_id, account_id, 1.0)

    async def test_account_deletion(self):
        group_id, transaction1_id = await self._create_group_with_transaction(
            "transfer"
        )
        account1_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account1",
            description="description",
        )

        # we can delete the account when nothing depends on it
        resp = await self._delete(f"/api/v1/accounts/{account1_id}")
        self.assertEqual(200, resp.status)

        account2_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account2",
            description="description",
        )

        await self._post_debitor_share(transaction1_id, account1_id, 1.0)
        await self._post_creditor_share(transaction1_id, account2_id, 1.0)
        await self._commit_transaction(transaction1_id)

        # we should not be able to delete this account as changes depend on it
        resp = await self._delete(f"/api/v1/accounts/{account2_id}")
        self.assertEqual(400, resp.status)

        await self._switch_creditor_share(transaction1_id, account1_id, 1.0)
        await self._commit_transaction(transaction1_id)

        # now we should be able to delete the account as nothing depends on it
        resp = await self._delete(f"/api/v1/accounts/{account2_id}")
        self.assertEqual(200, resp.status)

    async def test_purchase_items(self):
        group_id, transaction_id = await self._create_group_with_transaction("purchase")
        account1_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account1",
            description="foobar",
        )
        account2_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account2",
            description="foobar",
        )
        await self._post_debitor_share(transaction_id, account1_id, 1.0)
        await self._post_creditor_share(transaction_id, account2_id, 1.0)

        resp = await self._post(
            f"/api/v1/transactions/{transaction_id}/purchase_items",
            json={"name": "carrots", "price": 12.22, "communist_shares": 1},
        )
        self.assertEqual(200, resp.status)
        item_id = (await resp.json()).get("item_id")
        self.assertIsNotNone(item_id)

        # let's commit
        await self._commit_transaction(transaction_id)

        # now lets add some item shares, remove them again and commit
        resp = await self._post(
            f"/api/v1/purchase_items/{item_id}/shares",
            json={"account_id": account2_id, "share_amount": 1.0},
        )
        self.assertEqual(200, resp.status)

        resp = await self._post(
            f"/api/v1/purchase_items/{item_id}",
            json={"name": "carrots", "price": 12.22, "communist_shares": 0},
        )
        self.assertEqual(200, resp.status)
        resp = await self._delete(
            f"/api/v1/purchase_items/{item_id}/shares",
            json={
                "account_id": account2_id,
            },
        )
        self.assertEqual(200, resp.status)

        t = await self._fetch_transaction(transaction_id)
        self.assertIsNotNone(t["committed_positions"])
        self.assertEqual(1, len(t["committed_positions"]))
        self.assertEqual("carrots", t["committed_positions"][0]["name"])
        self.assertEqual(1, len(t["pending_positions"]))
        self.assertEqual(
            0,
            t["pending_positions"][0]["communist_shares"],
        )

        resp = await self._post(
            f"/api/v1/purchase_items/{item_id}/shares",
            json={"account_id": account2_id, "share_amount": 2.0},
        )
        self.assertEqual(200, resp.status)

        await self._commit_transaction(transaction_id=transaction_id)

        t = await self._fetch_transaction(transaction_id)
        self.assertIsNotNone(t["committed_positions"])
        self.assertEqual(1, len(t["committed_positions"]))
        self.assertIn(str(account2_id), t["committed_positions"][0]["usages"])

        resp = await self._delete(
            f"/api/v1/purchase_items/{item_id}",
        )
        self.assertEqual(200, resp.status)

        await self._commit_transaction(transaction_id=transaction_id)

        t = await self._fetch_transaction(transaction_id)
        self.assertIsNotNone(t["committed_positions"])
        self.assertEqual(1, len(t["committed_positions"]))
        self.assertTrue(t["committed_positions"][0]["deleted"])
